"""
Date:
Author:
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pickle
import random
import sys
import time
import numpy as np
from .svd import SVDModel

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.7'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiSVDppError(Exception):
    pass


class SVDpp(SVDModel):
    def __init__(self, matrix, k=1, matrix_p=None, matrix_q=None, name='SVDpp',
                 version=time.strftime("%Y%m%d", time.localtime())):
        """

        :param matrix:
        :param k:
        :param matrix_p:
        :param matrix_q:
        :param name:
        :param version:
        """
        assert matrix is not None, "'matrix' cannot be None."
        assert k > 0 and isinstance(k, int), "'k' should be an integer greater than 0."

        super().__init__(matrix=matrix, k=k, matrix_p=matrix_p, matrix_q=matrix_q, name=name, version=version)

    # TODO
    def train(self):
        pass

    # TODO
    def __fit(self):
        pass

    # TODO
    @staticmethod
    def restore(model_file_path):
        pass



"""Module for data preprocessing
Date: 25/May/2020
Author: Li Tang
"""
from __future__ import absolute_import, division, print_function, unicode_literals

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.9'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiToolboxPreprocessingError(Exception):
    pass


def month_to_int(month: str, unknown=None) -> int:
    """

    """
    month_dict = {'January': 1, 'February': 2, 'March': 3, 'April': 4, 'May': 5, 'June': 6, 'July': 7, 'August': 8,
                  'September': 9, 'October': 10, 'November': 11, 'December': 12}

    month_abbr_dict = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 'May': 5, 'Jun': 6, 'Jul': 7, 'Aug': 8, 'Sep': 9,
                       'Oct': 10, 'Nov': 11, 'Dec': 12}

    if month in month_dict:
        return month_dict[month]

    elif month in month_abbr_dict:
        return month_abbr_dict[month]

    else:
        if unknown is None:
            raise SuiToolboxPreprocessingError("This input month '{}' cannot be parsed.".format(month))
        else:
            return unknown


def weekday_to_int(weekday: str, unknown=None) -> int:
    weekday_dict = {'Monday': 1, 'Tuesday': 2, 'Wednesday': 3, 'Thursday': 4, 'Friday': 5, 'Saturday': 6, 'Sunday': 7}

    weekday_abbr_dict = {'Mon': 1, 'Tue': 2, 'Wed': 3, 'Thu': 4, 'Fri': 5, 'Sat': 6, 'Sun': 7}

    if weekday in weekday_dict:
        return weekday_dict[weekday]

    elif weekday in weekday_abbr_dict:
        return weekday_abbr_dict[weekday]

    else:
        if unknown is None:
            raise SuiToolboxPreprocessingError("This input weekday '{}' cannot be parsed.".format(weekday))
        else:
            return unknown



"""
    Module for data visualization
    Date: 25/May/2020
    Author: Li Tang
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from matplotlib import pyplot as plt
import pandas as pd
from sklearn.preprocessing import MinMaxScaler

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.6'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiDsVisualizationError(Exception):
    pass


def scatter(df, x, y, colour, title=None, x_label=None, y_label=None, shuffle=True, scaler=MinMaxScaler(),
            save_path=None):

    if scaler:
        df = pd.DataFrame(scaler.fit_transform(df[[x, y]].values), columns=[x, y])

    if shuffle:
        df = df.sample(frac=1)

    sc = plt.scatter(df[x], df[y], c=colour)
    plt.xlabel(x if x_label is None else x_label)
    plt.ylabel(y if y_label is None else y_label)
    plt.colorbar(sc)
    plt.subplots_adjust(top=2, bottom=1, right=2, left=1, hspace=0, wspace=0)
    if title is not None:
        plt.title(title, fontsize=14)
    if save_path is not None:
        plt.savefig(save_path, bbox_inches='tight', pad_inches=0, dpi=200)
    plt.show()



"""sui.ml
Machine learning algorithm implementations
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from .matrix_factorization import FunkSVD, BiasSVD, SVDpp, BPR, ALS

__all__ = ['FunkSVD', 'BiasSVD', 'SVDpp', 'BPR', 'ALS']



"""Module to recall based on BERT4Rec.
Date: 11/Jan/2021
Author: Li Tang
"""
__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.10'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'

import tensorflow as tf
from tensorflow.keras.layers import GRU
import os

import numpy as np
import matplotlib.pyplot as plt


import tensorflow_hub as hub
import tensorflow_datasets as tfds
tfds.disable_progress_bar()

from official.modeling import tf_utils
from official import nlp
from official.nlp import bert

# Load the required submodules
import official.nlp.optimization
import official.nlp.bert.bert_models
import official.nlp.bert.configs
import official.nlp.bert.run_classifier
import official.nlp.bert.tokenization
import official.nlp.data.classifier_data_lib
import official.nlp.modeling.losses
import official.nlp.modeling.models
import official.nlp.modeling.networks


class RNN4Rec:
    def __init__(self):
        pass


class GRU4Rec(GRU):
    def __init__(self, units, activation='tanh', recurrent_activation='sigmoid', use_bias=True,
                 kernel_initializer='glorot_uniform', recurrent_initializer='orthogonal', bias_initializer='zeros',
                 kernel_regularizer=None, recurrent_regularizer=None, bias_regularizer=None, activity_regularizer=None,
                 kernel_constraint=None, recurrent_constraint=None, bias_constraint=None, dropout=0.0,
                 recurrent_dropout=0.0, implementation=2, return_sequences=False, return_state=False,
                 go_backwards=False, stateful=False, unroll=False, time_major=False, reset_after=True, **kwargs):
        super().__init__(units, activation=activation, recurrent_activation=recurrent_activation, use_bias=use_bias,
                         kernel_initializer=kernel_initializer, recurrent_initializer=recurrent_initializer,
                         bias_initializer=bias_initializer, kernel_regularizer=kernel_regularizer,
                         recurrent_regularizer=recurrent_regularizer, bias_regularizer=bias_regularizer,
                         activity_regularizer=activity_regularizer, kernel_constraint=kernel_constraint,
                         recurrent_constraint=recurrent_constraint, bias_constraint=bias_constraint, dropout=dropout,
                         recurrent_dropout=recurrent_dropout, implementation=implementation,
                         return_sequences=return_sequences, return_state=return_state, go_backwards=go_backwards,
                         stateful=stateful, unroll=unroll, time_major=time_major, reset_after=reset_after, **kwargs)


class BERT4Rec(RNN4Rec):
    def __init__(self):
        super().__init__()


gs_folder_bert = "gs://cloud-tpu-checkpoints/bert/keras_bert/uncased_L-12_H-768_A-12"
tf.io.gfile.listdir(gs_folder_bert)



"""Attentional Factorization Machines
https://arxiv.org/abs/1708.04617
Date: 17/Aug/2020
Author: Li Tang
"""
import tensorflow as tf

import pickle
# from util.train_model import train_test_model_demo

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.2.0'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiAFMError(Exception):
    pass


class AttentionLayer(tf.keras.layers.Layer):
    def __init__(self, embedding_size=10, attention_size=3, **kwargs):
        self.embedding_size = embedding_size
        self.attention_size = attention_size
        super().__init__(**kwargs)

    def build(self, input_shape):
        input_dim = input_shape[2]

        self.linearlayer = tf.keras.layers.Dense(input_dim, activation='relu', use_bias=True)
        self.attention_w = self.add_weight(shape=(self.embedding_size, self.attention_size),
                                           initializer='random_normal', trainable=True)
        self.attention_b = self.add_weight(shape=(self.attention_size,),
                                           initializer='random_normal', trainable=True)
        self.attention_h = self.add_weight(shape=(self.attention_size,),
                                           initializer='random_normal', trainable=True)
        self.attention_p = self.add_weight(shape=(self.embedding_size, 1),
                                           initializer='ones', trainable=True)

    def call(self, input):
        # element_wise
        feature_dim = input.shape[1]
        element_wise_product_list = []
        for i in range(feature_dim):
            for j in range(i + 1, feature_dim):
                element_wise_product_list.append(tf.multiply(input[:, i, :], input[:, j, :]))  # None * embedding_size
        self.element_wise_product = tf.stack(element_wise_product_list)  # (F * F - 1 / 2) * None * embedding_size
        self.element_wise_product = tf.transpose(self.element_wise_product, perm=[1, 0, 2],
                                                 name='element_wise_product')  # None * (F * F - 1 / 2) *  embedding_size
        print("element_wise_product", self.element_wise_product.get_shape())
        # attention part
        num_interaction = int(feature_dim * (feature_dim - 1) / 2)
        # wx+b->relu(wx+b)->h*relu(wx+b)
        self.attention_wx_plus_b = tf.reshape(
            tf.add(tf.matmul(tf.reshape(self.element_wise_product, shape=(-1, self.embedding_size)),
                             self.attention_w), self.attention_b),
            shape=[-1, num_interaction, self.attention_size])  # N *(F*F-1/2)*1
        self.attention_exp = tf.exp(tf.reduce_sum(tf.multiply(tf.nn.relu(self.attention_wx_plus_b),
                                                              self.attention_h), axis=2))  # N * ( F * F - 1 / 2) * 1

        self.attention_exp_sum = tf.reshape(tf.reduce_sum(self.attention_exp, axis=1), shape=(-1, 1))  # N * 1 * 1

        self.attention_out = tf.divide(self.attention_exp, self.attention_exp_sum,
                                       name='attention_out')  # N * ( F * F - 1 / 2) * 1
        self.attention_x_product = tf.reduce_sum(
            tf.einsum('bn,bnm->bnm', self.attention_out, self.element_wise_product), axis=1,
            name='afm')  # N * embedding_size
        self.attention_part_sum = tf.matmul(self.attention_x_product, self.attention_p)  # N * 1

        return self.attention_part_sum


class AFM(tf.keras.Model):
    def __init__(self, feature_dim, embedding_size=10, attention_size=3):
        super().__init__()
        self.feature_dim = feature_dim  # F features nums 字典数量
        self.embedding_size = embedding_size
        self.attention_size = attention_size
        # Embedding 这里采用embeddings层 因此大小为F* M F为field特征数量，N 为 feature的种类数 M为embedding的维度
        feat_embeddings = tf.keras.layers.Embedding(feature_dim, embedding_size,
                                                    embeddings_initializer='uniform')  # N * embedding_size
        self.feat_embeddings = feat_embeddings
        self.attentionlayer = AttentionLayer(self.embedding_size, self.attention_size)
        # linear part
        self.linearlayer = tf.keras.layers.Dense(1, activation='relu', use_bias=True)

    def call(self, feat_index, feat_value):
        # call函数接收输入变量
        # embedding part  feat_index = inputs为输入 feat_embeddings为一个layer。
        feat_embedding_0 = self.feat_embeddings(feat_index)  # Batch * F * embedding_size
        feat_embedding = tf.einsum('bnm,bn->bnm', feat_embedding_0, feat_value)  # # Batch * F * embedding_size
        feat_embedding_1 = tf.transpose(feat_embedding, perm=[0, 2, 1])
        y_deep = self.attentionlayer(feat_embedding)

        y_linear = tf.reduce_sum(self.linearlayer(feat_embedding_1), axis=1)
        output = y_deep + y_linear
        return output



"""Module including some implementations of matrix factorization
Date: 28/Mar/2019
Author: Li Tang
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import json
import pickle
import time
from sui.toolbox import top_k
import numpy as np

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.7'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiMlMfError(Exception):
    pass


class SVDModel:
    def __init__(self, matrix, k=1, matrix_p=None, matrix_q=None, name='SVD',
                 version=time.strftime("%Y%m%d", time.localtime())):
        """

        :param matrix:
        :param k:
        :param matrix_p:
        :param matrix_q:
        :param name:
        :param version:
        """
        assert matrix is not None, "'matrix' cannot be None."
        assert k > 0 and isinstance(k, int), "'k' should be an integer greater than 0."

        self._matrix = None
        self._matrix_min = None
        self._matrix_max = None
        self.set_matrix(matrix)
        self.k = k
        self.matrix_p = np.random.rand(len(self._matrix), self.k) if matrix_p is None else matrix_p
        self.matrix_q = np.random.rand(self.k, len(self._matrix[0])) if matrix_q is None else matrix_q
        self.name = name
        self.version = version

    def set_matrix(self, matrix):
        matrix = np.array(matrix, dtype=np.float64)
        self._matrix_min = np.nanmin(matrix)
        self._matrix_max = np.nanmax(matrix)
        self._matrix = (matrix - self._matrix_min) / (self._matrix_max - self._matrix_min)

    def get_matrix(self):
        print(self._matrix)
        return self._matrix * (self._matrix_max - self._matrix_min) + self._matrix_min

    def train(self):
        raise NotImplementedError

    def predict(self, topk=1, target='p', result_path=None, only_vacancy=True):
        """

        :param topk:
        :param target:
        :param result_path:
        :param only_vacancy:
        :return:
        """
        result_dict = {}
        if target == 'p':
            for row in range(len(self._matrix)):
                topk_reco = []
                for col in range(len(self._matrix[row])):
                    if self._matrix[row, col] is None or np.isnan(self._matrix[row, col]) or only_vacancy is False:
                        score = np.matmul(self.matrix_p[row, :], self.matrix_q.T[col, :])
                        topk_reco.append([col, score])
                result_dict[row] = top_k(data=topk_reco, k=topk, axis=1, desc=True)
        elif target == 'q':
            for col in range(len(self._matrix[0])):
                topk_reco = []
                for row in range(len(self._matrix)):
                    if self._matrix[row, col] is None or np.isnan(self._matrix[row, col]):
                        score = np.matmul(self.matrix_p[row, :], self.matrix_q.T[col, :])
                        topk_reco.append([row, score])
                result_dict[col] = top_k(data=topk_reco, k=topk, axis=1, desc=True)
        else:
            raise SuiMlMfError("'target' should be 'p' or 'q'. Obtained:", target)

        if result_path is not None:
            result_file = '{}_{}.json'.format(self.name, self.version)
            try:
                with open(result_path + result_file, 'w') as result_output:
                    json.dump([result_dict], result_output)
                print('Result is saved into {}.'.format(result_path + result_file))
            except Exception as e:
                raise SuiMlMfError('Failed to save result:', e)

        return result_dict

    def dump(self, model_file_path=None):
        """

        :param model_file_path: Absolute path to dump this model via using pickle.
        :return:
        """
        if model_file_path is None:
            model_file_path = './{}_{}.pkl'.format(self.name, self.version)
        try:
            with open(model_file_path, 'wb') as model_output:
                model_info = dict()
                model_info['matrix'] = self.get_matrix()
                model_info['name'] = self.name
                model_info['version'] = self.version
                model_info['k'] = self.k
                model_info['matrix_p'] = self.matrix_p
                model_info['matrix_q'] = self.matrix_q
                pickle.dump(model_info, model_output)

            print('Model is saved into {}.'.format(model_file_path))

        except Exception as e:
            raise SuiMlMfError('Failed to dump model:', e)

    @staticmethod
    def restore(model_file_path):
        raise NotImplementedError



"""sui.dl
Deep learning algorithm implementations
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from .afm import AFM
# from .bert4rec import BERT4Rec
from .pnn import PNN

# __all__ = ['AFM', 'BERT4Rec', 'PNN']
# __all__ = ['AFM', 'PNN']



"""Product-based Neural Networks
https://arxiv.org/pdf/1611.00144.pdf
Date: 14/Jul/2020
Author: Li Tang
"""
import tensorflow as tf
from tensorflow.keras.layers import Activation, BatchNormalization, Dense, Dropout
import numpy as np
from .initializers import get_init
from .losses import get_loss
from .optimizers import get_opti

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.2.0'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiPNNError(Exception):
    pass


class PNN(tf.keras.Model):
    def __init__(self, features_dim: int, fields_dim: int, hidden_layer_sizes: list, dropout_params: list,
                 product_layer_dim=10, lasso=0.01, ridge=1e-5, embedding_dim=10, product_type='pnn',
                 initializer='glorotuniform', activation='sigmoid', hidden_activation='relu'):
        super().__init__()
        self.features_dim = features_dim  # size of features after one-hot, denoted by F
        self.fields_dim = fields_dim  # number of different original features, denoted by N
        self.dropout_params = dropout_params
        self.hidden_layer_sizes = hidden_layer_sizes  # number of hidden layers
        self.product_layer_dim = product_layer_dim  # as same as the input size of l_1, denoted by D_1
        self.lasso = lasso
        self.ridge = ridge
        self.embedding_dim = embedding_dim  # dimension of vectors after embedding, denoted by M
        # product type for product layer
        # 'ipnn' for inner product , 'opnn' for outer product, and 'pnn' for concatenating both product
        self.product_type = product_type
        self.initializer = get_init(initializer)
        self.activation = activation
        self.hidden_activation = hidden_activation

        # embedding layer
        # the size of embedding layer is F * M
        self.embedding_layer = tf.keras.layers.Embedding(self.features_dim, self.embedding_dim,
                                                         embeddings_initializer='uniform')

        # product layer
        # linear signals l_z
        self.linear_sigals_variable = tf.Variable(
            self.initializer(shape=(self.product_layer_dim, self.fields_dim, self.embedding_dim)))
        # quadratic signals l_p
        self.__init_quadratic_signals()

        # hidden layers
        self.__init_hidden_layers()

        # output layer
        self.output_layer = tf.keras.layers.Dense(1, activation=self.activation, use_bias=True)

    def __init_quadratic_signals(self):
        if self.product_type == 'ipnn':
            # matrix decomposition based on the assumption: W_p^n = \theta ^n * {\theta^n}^T
            # then the size of W_p^n is D_1 * N
            self.theta = tf.Variable(self.initializer(shape=(self.product_layer_dim, self.fields_dim)))
        elif self.product_type == 'opnn':
            # the size of W_p^n is D_1 * M * M
            self.quadratic_weights = tf.Variable(
                self.initializer(shape=(self.product_layer_dim, self.embedding_dim, self.embedding_dim)))
        elif self.product_type == 'pnn':
            self.theta = tf.Variable(self.initializer(shape=(self.product_layer_dim, self.fields_dim)))
            self.quadratic_weights = tf.Variable(
                self.initializer(shape=(self.product_layer_dim, self.embedding_dim, self.embedding_dim)))
        else:
            raise SuiPNNError("'product_type' should be 'ipnn', 'opnn', or 'pnn'.")

    def __init_hidden_layers(self):
        for layer_index in range(len(self.hidden_layer_sizes)):
            setattr(self, 'dense_' + str(layer_index), Dense(self.hidden_layer_sizes[layer_index]))
            setattr(self, 'batch_norm_' + str(layer_index), BatchNormalization())
            setattr(self, 'activation_' + str(layer_index), Activation(self.hidden_activation))
            setattr(self, 'dropout_' + str(layer_index), Dropout(self.dropout_params[layer_index]))

    def call(self, feature_value, embedding_index, training=False):
        features = tf.einsum('bnm,bn->bnm', self.embedding_layer(embedding_index), feature_value)
        # linear part
        l_z = tf.einsum('bnm,dnm->bd', features, self.linear_sigals_variable)  # Batch * D_1

        # quadratic part
        if self.product_type == 'ipnn':
            delta = tf.einsum('dn,bnm->bdnm', self.theta, features)  # Batch * D_1 * N * M
            l_p = tf.einsum('bdnm,bdnm->bd', delta, delta)
        elif self.product_type == 'opnn':
            sum_features = tf.einsum('bnm->bm', features)  # Batch * M
            p = tf.einsum('bm,bn->bmn', sum_features, sum_features)
            l_p = tf.einsum('bmn,dmn->bd', p, self.quadratic_weights)
        elif self.product_type == 'pnn':
            delta = tf.einsum('dn,bnm->bdnm', self.theta, features)  # Batch * D_1 * N * M
            sum_features = tf.einsum('bnm->bm', features)  # Batch * M
            p = tf.einsum('bm,bn->bmn', sum_features, sum_features)
            l_p = tf.concat(
                (tf.einsum('bdnm,bdnm->bd', delta, delta), tf.einsum('bmn,dmn->bd', p, self.quadratic_weights)), axis=1)
        else:
            raise SuiPNNError("'product_type' should be 'ipnn', 'opnn', or 'pnn'.")

        model = tf.concat((l_z, l_p), axis=1)
        if training:
            model = tf.keras.layers.Dropout(self.dropout_params[0])(model)

        for i in range(len(self.hidden_layer_sizes)):
            model = getattr(self, 'dense_' + str(i))(model)
            model = getattr(self, 'batch_norm_' + str(i))(model)
            model = getattr(self, 'activation_' + str(i))(model)
            if training:
                model = getattr(self, 'dropout_' + str(i))(model)

        return self.output_layer(model)

    def train(self, feature_value, embedding_index, label, optimizer='adam', learning_rate=1e-4, loss='sigmoid',
              epochs=50, batch=32, shuffle=10000):
        for epoch in range(epochs):
            train_set = tf.data.Dataset.from_tensor_slices((feature_value, embedding_index, label)).shuffle(
                shuffle).batch(batch, drop_remainder=True)
            for batch_set in train_set:
                with tf.GradientTape() as tape:
                    prediction = self.call(feature_value=batch_set[0], embedding_index=batch_set[1], training=True)
                    self.loss_obj = get_loss(loss)
                    self.optimizer = get_opti(optimizer)(learning_rate=learning_rate)
                    batch_loss = self.loss_obj(batch_set[2], prediction)
                gradients = tape.gradient(batch_loss, self.trainable_variables)
                self.optimizer.apply_gradients(zip(gradients, self.trainable_variables))
            mean_loss = tf.keras.metrics.Mean(name='train_loss')
            print('epoch: {} ==> loss: {}'.format(epoch + 1, mean_loss(batch_loss)))

    def predict(self, feature_value, embedding_index):
        feature_value = tf.convert_to_tensor(feature_value)
        embedding_index = tf.convert_to_tensor(embedding_index)
        return self.call(feature_value=feature_value, embedding_index=embedding_index, training=False)

    # TODO
    def dump(self, path):
        self.save(filepath=path)

    # TODO
    @staticmethod
    def restore():
        return None



"""sui.ml.matrix_factorization
Machine learning algorithms about matrix factorization
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from .funksvd import FunkSVD
from .biassvd import BiasSVD
from .svdpp import SVDpp
from .bpr import BPR
from .als import ALS

__all__ = ['FunkSVD', 'BiasSVD', 'SVDpp', 'BPR', 'ALS']



import tensorflow as tf


class SuiOptimizersError(ValueError):
    pass


def get_opti(opti_name: str):
    if opti_name == 'adam':
        return tf.keras.optimizers.Adam
    else:
        raise SuiOptimizersError



"""Convert user's log data into series data
Date: 02/Mar/2021
Author: Li Tang
"""
from typing import Union
import random

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.11'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


def build_session(data: Union[list, tuple], user_idx: Union[int, None], item_idx: int, timestamp_idx: int, session_gap: int):
    """Function to split data into multiple sessions

    Args:
        data: user's log data sorted input data consists of lists or tuples, including user id, item id, and timestamp
        user_idx: index of the user id in each user's log data; set user_idx to None can build session without user id
        item_idx: index of the item id in each user's log data
        timestamp_idx: index of the timestamp in each user's log data
        session_gap: the threshold to split user's activities into multiple sessions based the gap of adjacent timestamps

    Returns:
        a list sorted by target dimension with length no greater than k

    Examples:
        >>> log_data = (('user_867as8e', 'v9d8cv8272lk', 1614652639000),
        ...             ('user_867as8e', 'v8d4ln9834kj', 1614653499000),
        ...             ('user_868yu82', 'n0s3mn43k4n3', 1614653646000),
        ...             ('user_868yu82', 'v987d3n5l89n', 1614653702246))
        >>> build_session(data=log_data, user_idx=0, item_idx=1, timestamp_idx=2, session_gap=600000)
        [('user_867as8e', ['v9d8cv8272lk']), ('user_867as8e', ['v8d4ln9834kj']), ('user_868yu82', ['n0s3mn43k4n3', 'v987d3n5l89n'])]
        >>> log_data = (('v9d8cv8272lk', 1614652639000),
        ...             ('v8d4ln9834kj', 1614653499000),
        ...             ('n0s3mn43k4n3', 1614653646000),
        ...             ('v987d3n5l89n', 1614653702246))
        >>> build_session(data=log_data, user_idx=None, item_idx=0, timestamp_idx=1, session_gap=100000)
        [['v9d8cv8272lk'], ['v8d4ln9834kj'], ['n0s3mn43k4n3', 'v987d3n5l89n']]

    """
    is_first_log = True
    result = []
    current_ts = 0
    session = []

    # if user id exists
    if user_idx is not None:
        user_id = ''
        for log in data:
            # if this is the first log
            if is_first_log:
                user_id = log[user_idx]
                current_ts = log[timestamp_idx]
                is_first_log = False

            if user_id == log[user_idx] and log[timestamp_idx] - current_ts <= session_gap:
                session.append(log[item_idx])
            else:
                result.append((user_id, session))
                session = [log[item_idx]]
                user_id = log[user_idx]
                current_ts = log[timestamp_idx]
        result.append((user_id, session))

        return result

    # else build sessions with no user id
    else:
        for log in data:
            if is_first_log:
                current_ts = log[timestamp_idx]
                is_first_log = False

            if log[timestamp_idx] - current_ts <= session_gap:
                session.append(log[item_idx])
            else:
                result.append(session)
                session = [log[item_idx]]
                current_ts = log[timestamp_idx]
        result.append(session)

        return result


def random_walk(G, walk_depth: int, walk_path: list) -> list:
    """Function to generate a series data by walking in the input graph data randomly

    Args:
        G: input graph data
        walk_depth: the length of the walking result
        walk_path: the existing walking history; walking will start at the last element in walk_path

    Returns:
        a list consisting of every passed vertex

    Examples:
        >>> import networkx as nx
        >>> G = nx.Graph()
        >>> G.add_nodes_from([
        ...     ('Beijing', {'Country': 'CN'}),
        ...     ('London', {'Country': 'UK'}),
        ...     ('Zhuzhou', {'Country': 'CN'}),
        ...     ('Manchester', {'Country': 'UK'}),
        ...     ('New York', {'Country': 'US'}),
        ...     ('Michigan', {'Country': 'US'}),
        ...     ('Shanghai', {'Country': 'CN'})
        ... ])
        >>> G.add_edges_from([
        ...     ('Beijing', 'Zhuzhou'),
        ...     ('Beijing', 'London'),
        ...     ('Beijing', 'New York'),
        ...     ('Beijing', 'Shanghai'),
        ...     ('Zhuzhou', 'Shanghai'),
        ...     ('Shanghai', 'London'),
        ...     ('Shanghai', 'New York'),
        ...     ('Shanghai', 'Michigan'),
        ...     ('London', 'Manchester'),
        ...     ('London', 'New York'),
        ...     ('New York', 'Michigan')
        ... ])
        >>> result = []
        >>> for vertex in G.nodes:
        ...     result.append(random_walk(G=G, walk_depth=5, walk_path=[vertex]))
        >>> len(result), len(result[0])
        (7, 5)
        >>> [_[0] for _ in result]
        ['Beijing', 'London', 'Zhuzhou', 'Manchester', 'New York', 'Michigan', 'Shanghai']

    """
    if len(walk_path) < 1:
        raise ValueError("'walk_path' should include at least 1 vertex to start walking.")
    while len(walk_path) < walk_depth:
        current_vertex = walk_path[-1]
        neighbors_list = list(G.neighbors(current_vertex))
        if len(neighbors_list) > 0:
            walk_path.append(random.choice(neighbors_list))
        else:
            break

    return walk_path



"""Top k function for lists or tuples
Date: 26/Feb/2021
Author: Li Tang
"""
from typing import Union, List

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.11'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


def top_k(data: Union[list, tuple], k: int, axis: int = 1, target: int = 0, target_only: bool = False,
          desc: bool = True) -> List[list or tuple]:
    """Function to return the top k values in the input list/tuple

    Args:
        data: input data to be sorted consists of lists or tuples
        k: the length of result to return
        axis: the dimension in the list/tuple in data for sorting
        target: the target dimension to return after sorting
        target_only: if target_only is True, the result contains only the target dimension in data;
        otherwise return all dimensions in data
        desc: True for descending and False for ascending

    Returns:
        a list sorted by target dimension with length no greater than k

    Examples:
        >>> data = (('Beijing', 8610, 'CN'), ('London', 4420, 'UK'), ('Zhuzhou', 86731, 'CN'), ('Manchester', 44121, 'UK'))
        >>> top_k(data=data, k=3, axis=1, target=0, target_only=False, desc=True)
        [('Zhuzhou', 86731, 'CN'), ('Manchester', 44121, 'UK'), ('Beijing', 8610, 'CN')]

    """

    def __push(item, data, axis, desc):
        for idx in range(len(data) - 1, -1, -1):
            if desc is True and data[idx][axis] >= item[axis] or desc is False and data[idx][axis] <= item[axis]:
                result = data[:idx + 1]
                result.append(item)
                result.extend(data[idx + 1:])
                return result
        data.insert(0, item)
        return data

    result = []

    for item in data:
        if len(result) < k or desc is True and item[axis] > result[-1][axis] or desc is False and item[axis] < \
                result[-1][axis]:
            result = __push(item, result, axis=axis, desc=desc)
        if len(result) > k:
            result = result[:k]

    return [i[target] for i in result] if target_only else result



from __future__ import absolute_import, division, print_function, unicode_literals
from .ml import FunkSVD, BiasSVD, SVDpp, BPR, ALS
from .dl import AFM, PNN
from .graph import DeepWalk

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.2.0'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'

my_dear = "Dear Miss Sui Lin, I love you!"

__api_info_dict = {
    "sui.ALS": "Alternating Least Squares",
    "sui.BiasSVD": "BiasSVD",
    "sui.BPR": "Bayesian Personalized Ranking",
    "sui.FunkSVD": "FunkSVD",
    "sui.GRU4Rec": "GRU4Rec",
    "sui.PNN": "Product-based Neural Networks",
    "sui.SVDpp": "SVD++",
    "sui.DeepWalk": "A deepwalk model to embed nodes vectors from a given graph",
    "sui.toolbox.top_k": "Return a list containing top k data for a specific dimension",
    "sui.toolbox.random_walk": "Generate a list including walking paths based on random walk in the input graph"
}


def api_info(api=None):
    if api is not None:
        if api in __api_info_dict:
            print('API: {}\nInfo: {}\n'.format(api, __api_info_dict[api]))
        else:
            print('{} is not a correct API.'.format(api))
    else:
        for api, info in __api_info_dict.items():
            print('API: {}\nInfo: {}\n'.format(api, info))



"""Module including some implementations of matrix factorization
Date: 28/Mar/2019
Author: Li Tang
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pickle
import random
import sys
import time
import numpy as np
from .svd import SVDModel

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.7'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiFunkSVDError(Exception):
    pass


class FunkSVD(SVDModel):
    def __init__(self, matrix, k=1, matrix_p=None, matrix_q=None, name='FunkSVD',
                 version=time.strftime("%Y%m%d", time.localtime())):
        """

        :param matrix:
        :param k:
        :param matrix_p:
        :param matrix_q:
        :param name:
        :param version:
        """
        assert matrix is not None, "'matrix' cannot be None."
        assert k > 0 and isinstance(k, int), "'k' should be an integer greater than 0."

        super().__init__(matrix=matrix, k=k, matrix_p=matrix_p, matrix_q=matrix_q, name=name, version=version)

    def train(self, penalty='ridge', penalty_weight=0.5, learning_rate=0.75, learning_rate_decay=1.0,
              min_learning_rate=None, dropout=0.0, epochs=50, early_stopping=10, workers=1):
        """
        :param penalty: penalty can be either lasso or ridge; hybrid of them is not allowed
        :param penalty_weight: weight of the given type of penalty
        :param learning_rate: initial learning rate
        :param learning_rate_decay: decay
        :param min_learning_rate:
        :param dropout:
        :param epochs:
        :param early_stopping:
        :param workers:
        :return:
        """
        assert penalty in ['ridge', 'lasso'], "'penalty' should be either 'ridge' or 'lasso'."
        assert penalty_weight > 0, "'penalty_weight' should be greater than 0."
        assert learning_rate > 0, "'learning_rate' should be greater than 0."
        assert learning_rate_decay > 0, "'learning_rate_decay' should be greater than 0." \
                                        " Set 1 for no decay in training."
        assert 0 <= dropout < 1.0, "The domain of definition of 'dropout' should be [0, 1)."
        assert isinstance(early_stopping, int), \
            "'early_stopping' should be an integer. Set 0 or any negative integer to interdict early stopping."
        assert isinstance(workers, int) and workers > 0, "'workers' should be an integer greater than 0."

        learning_rate /= 1 - dropout

        loss_history = [sys.maxsize]
        for epoch in range(epochs):
            loss = 0
            trained_samples = 0
            skipped_samples = 0

            if min_learning_rate:
                if learning_rate < min_learning_rate:
                    learning_rate = min_learning_rate

            # TODO: split matrix into multiple sections for concurrency
            row_purview_list = [(0, len(self._matrix))]
            col_purview_list = [(0, len(self._matrix[0]))]

            # start to train
            for section_idx in range(len(row_purview_list)):
                section_loss, section_trained_samples, section_skipped_samples = self.__fit(
                    row_purview_list[section_idx], col_purview_list[section_idx], learning_rate, penalty,
                    penalty_weight, dropout)
                loss += section_loss
                trained_samples += section_trained_samples
                skipped_samples += section_skipped_samples

            print('epoch: {} ==> loss: {}'.format(epoch + 1, loss))

            if dropout > 0:
                print('Trained {} samples and skipped {} samples.'
                      ' The dropout rate is: {}%'.format(trained_samples, skipped_samples, round(
                    skipped_samples / (trained_samples + skipped_samples) * 100)))

            if learning_rate_decay != 1.0:
                print('Current learning rate: {}'.format(learning_rate))
                learning_rate *= learning_rate_decay

            if early_stopping > 0:
                if loss < loss_history[0]:
                    loss_history = [loss]
                else:
                    loss_history.append(loss)

                if len(loss_history) >= early_stopping:
                    print(
                        'Early stopping! The best performance is at No.{} epoch and the loss have not been decreased'
                        ' from then on as {}:'.format(epoch - early_stopping + 2, loss_history))
                    break
            else:
                continue

    def __fit(self, row_purview: tuple, col_purview: tuple, learning_rate, penalty, penalty_weight, dropout):
        """

        :param row_purview:
        :param col_purview:
        :param learning_rate:
        :param penalty:
        :param penalty_weight:
        :param dropout:
        :return:
        """
        loss = 0
        trained_samples = 0
        skipped_samples = 0

        for row in range(row_purview[0], row_purview[1]):
            for col in range(col_purview[0], col_purview[1]):
                if self._matrix[row, col] is None or np.isnan(self._matrix[row, col]):
                    continue
                if random.random() <= 1 - dropout:
                    y_hat = np.matmul(self.matrix_p[row, :], self.matrix_q.T[col, :])

                    if penalty == 'ridge':
                        self.matrix_p[row, :] += learning_rate * ((self._matrix[row, col] - y_hat) *
                                                                  self.matrix_q[:, col] - penalty_weight *
                                                                  self.matrix_p[row, :]) / self.k

                        self.matrix_q[:, col] += learning_rate * ((self._matrix[row, col] - y_hat) *
                                                                  self.matrix_p[row, :] - penalty_weight *
                                                                  self.matrix_q[:, col]) / self.k

                        loss += ((self._matrix[row, col] - y_hat) ** 2 + penalty_weight * (
                                np.linalg.norm(self.matrix_p[row, :]) + np.linalg.norm(
                            self.matrix_q.T[col, :]))) / self.k
                    elif penalty == 'lasso':
                        self.matrix_p[row, :] += learning_rate * ((self._matrix[row, col] - y_hat) *
                                                                  self.matrix_q[:, col] - penalty_weight) / self.k
                        self.matrix_q[:, col] += learning_rate * ((self._matrix[row, col] - y_hat) *
                                                                  self.matrix_p[row, :] - penalty_weight) / self.k
                        loss += ((self._matrix[row, col] - y_hat) ** 2 + penalty_weight * (
                                np.linalg.norm(self.matrix_p[row, :], ord=1) + np.linalg.norm(
                            self.matrix_q.T[col, :], ord=1))) / self.k
                    else:
                        raise ValueError
                    trained_samples += 1
                else:
                    skipped_samples += 1

        return loss, trained_samples, skipped_samples

    @staticmethod
    def restore(model_file_path):
        """

        :param model_file_path:
        :return:
        """
        try:
            with open(model_file_path, 'rb') as model_input:
                model_info = pickle.load(model_input)
                model = FunkSVD(**model_info)
            print('Model is loaded from {}.'.format(model_file_path))
            return model

        except Exception as e:
            raise SuiFunkSVDError('Failed to restore model:', e)

    # TODO: add new user or product into a pretrained matrix
    def add(self, target, value, initializer='mean'):
        """

        :param target:
        :param value:
        :param initializer:
        :return:
        """
        assert target in ['k', 'matrix_p', 'matrix_q'], "'target' cannot be found."
        assert isinstance(value, int) and value > 0, "'value' should be an integer greater than 0."
        assert initializer in ['mean', 'random'], "'initializer' should be either 'mean' or 'random'."



import tensorflow as tf


class SuiLossesError(ValueError):
    pass


def get_loss(loss_name: str):
    if loss_name == 'sigmoid':
        return tf.keras.losses.BinaryCrossentropy
    elif loss_name == 'mse':
        return tf.keras.losses.MSE
    else:
        raise SuiLossesError



import tensorflow as tf

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.2.0'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiInitializersError(ValueError):
    pass


def get_init(initializer_name: str):
    if initializer_name == 'glorotnormal':
        return tf.initializers.GlorotNormal
    elif initializer_name == 'glorotuniform':
        return tf.initializers.GlorotUniform()
    else:
        raise SuiInitializersError



"""DeepWalk class to initiate, train, and use a deepwalk model
described in https://arxiv.org/pdf/1403.6652.pdf
based on networkx and gensim

Date: 07/Mar/2021
Author: Li Tang
"""
from typing import Union, List, Tuple
from multiprocessing import Pool

import networkx as nx
from gensim.models import Word2Vec

from ..toolbox import random_walk

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.11'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


class SuiDeepWalkError(Exception):
    pass


class DeepWalk:
    """Initiate, train and use a deepwalk model described in https://arxiv.org/pdf/1403.6652.pdf;
    this class is implemented based on gensim and networkx.

    """

    def __init__(self, G=nx.Graph(), model=None, walk_path: list = []):
        self.G = G  # networkx graph to generate the walk path to train the embedding model
        self.model = model  # gensim word2vec model to embed node vectors
        self.walk_path = walk_path  # walk path to train the model as 'sentences'

    def add_node(self, node_for_adding):
        self.G.add_node(node_for_adding=node_for_adding)

    def add_nodes_from(self, nodes_for_adding):
        self.G.add_nodes_from(nodes_for_adding=nodes_for_adding)

    def add_edge(self, u_of_edge, v_of_edge):
        self.G.add_edge(u_of_edge=u_of_edge, v_of_edge=v_of_edge)

    def add_edges_from(self, ebunch_to_add):
        self.G.add_edges_from(ebunch_to_add=ebunch_to_add)

    def concurrent_walk(self, walk_depth: int, walking_workers: int = 4, nodes_list: List[list] = None) -> List[list]:
        """Function to obtain the series vertex data by concurrent walking in the graph.

        Args:
        walk_depth: the length for each walking result.
        walking_workers: the maximum number of processes in the pool.
        nodes_list: the list containing existing walking history;
            the last element in this list would be considered as the start for walking;
            if the nodes_list is None, every vertex in G.nodes will combine the nodes_list.

        Returns:
            a list including lists of walking series which consist of passed nodes.

        Examples:
            >>> G = nx.Graph()
            >>> G.add_nodes_from([
            ...     ('Beijing', {'Country': 'CN'}),
            ...     ('London', {'Country': 'UK'}),
            ...     ('Zhuzhou', {'Country': 'CN'}),
            ...     ('Manchester', {'Country': 'UK'}),
            ...     ('New York', {'Country': 'US'}),
            ...     ('Michigan', {'Country': 'US'}),
            ...     ('Shanghai', {'Country': 'CN'})
            ... ])
            >>> G.add_edges_from([
            ...     ('Beijing', 'Zhuzhou'),
            ...     ('Beijing', 'London'),
            ...     ('Beijing', 'New York'),
            ...     ('Beijing', 'Shanghai'),
            ...     ('Zhuzhou', 'Shanghai'),
            ...     ('Shanghai', 'London'),
            ...     ('Shanghai', 'New York'),
            ...     ('Shanghai', 'Michigan'),
            ...     ('London', 'Manchester'),
            ...     ('London', 'New York'),
            ...     ('New York', 'Michigan')
            ... ])
            >>> walker = DeepWalk(G=G)
            >>> result_list = walker.concurrent_walk(5, 2)
            >>> [_[0] for _ in result_list]
            ['Beijing', 'London', 'Zhuzhou', 'Manchester', 'New York', 'Michigan', 'Shanghai']
            >>> new_result_list = walker.concurrent_walk(8, 2, nodes_list=result_list)
            >>> len(new_result_list), len(new_result_list[0])
            (7, 8)
            >>> [result_list[i][:5] == new_result_list[i][:5] for i in range(len(result_list))]
            [True, True, True, True, True, True, True]

        """
        # if there is no specific nodes list to generate the walk path
        # uses all nodes in the graph directly
        if nodes_list is None:
            if len(self.G.nodes) == 0:
                raise ValueError("Parameter 'nodes_list' cannot be None if the length of G.nodes is 0.")

            nodes_list = [[node] for node in self.G.nodes]

        # multiprocess pool to schedule the concurrent random walk
        pool = Pool(walking_workers)
        result = []

        for walk_path in nodes_list:
            result.append(pool.apply_async(random_walk, args=(self.G, walk_depth, walk_path)))

        pool.close()
        pool.join()

        return [_.get() for _ in result]

    def _embedding(self, sentences=None, size=100, alpha=0.025, window=5, min_count=5,
                   max_vocab_size=None, sample=1e-3, seed=1, workers=3, min_alpha=0.0001,
                   sg=0, hs=0, negative=5, ns_exponent=0.75, cbow_mean=1, hashfxn=hash, epochs=5, null_word=0,
                   trim_rule=None, sorted_vocab=1, batch_words=10000, compute_loss=False, callbacks=(),
                   max_final_vocab=None):
        """Function to train the word2vec model based on gensim.
        Args docstring are copied from gensim.models.Word2Vec.

        Args:
        sentences : iterable of iterables, optional
            The `sentences` iterable can be simply a list of lists of tokens, but for larger corpora,
            consider an iterable that streams the sentences directly from disk/network.
            See :class:`~gensim.models.word2vec.BrownCorpus`, :class:`~gensim.models.word2vec.Text8Corpus`
            or :class:`~gensim.models.word2vec.LineSentence` in :mod:`~gensim.models.word2vec` module for such examples.
            See also the `tutorial on data streaming in Python
            <https://rare-technologies.com/data-streaming-in-python-generators-iterators-iterables/>`_.
            If you don't supply `sentences`, the model is left uninitialized -- use if you plan to initialize it
            in some other way.
        corpus_file : str, optional
            Path to a corpus file in :class:`~gensim.models.word2vec.LineSentence` format.
            You may use this argument instead of `sentences` to get performance boost. Only one of `sentences` or
            `corpus_file` arguments need to be passed (or none of them, in that case, the model is left uninitialized).
        size : int, optional
            Dimensionality of the word vectors.
        window : int, optional
            Maximum distance between the current and predicted word within a sentence.
        min_count : int, optional
            Ignores all words with total frequency lower than this.
        workers : int, optional
            Use these many worker threads to train the model (=faster training with multicore machines).
        sg : {0, 1}, optional
            Training algorithm: 1 for skip-gram; otherwise CBOW.
        hs : {0, 1}, optional
            If 1, hierarchical softmax will be used for model training.
            If 0, and `negative` is non-zero, negative sampling will be used.
        negative : int, optional
            If > 0, negative sampling will be used, the int for negative specifies how many "noise words"
            should be drawn (usually between 5-20).
            If set to 0, no negative sampling is used.
        ns_exponent : float, optional
            The exponent used to shape the negative sampling distribution. A value of 1.0 samples exactly in proportion
            to the frequencies, 0.0 samples all words equally, while a negative value samples low-frequency words more
            than high-frequency words. The popular default value of 0.75 was chosen by the original Word2Vec paper.
            More recently, in https://arxiv.org/abs/1804.04212, Caselles-Dupré, Lesaint, & Royo-Letelier suggest that
            other values may perform better for recommendation applications.
        cbow_mean : {0, 1}, optional
            If 0, use the sum of the context word vectors. If 1, use the mean, only applies when cbow is used.
        alpha : float, optional
            The initial learning rate.
        min_alpha : float, optional
            Learning rate will linearly drop to `min_alpha` as training progresses.
        seed : int, optional
            Seed for the random number generator. Initial vectors for each word are seeded with a hash of
            the concatenation of word + `str(seed)`. Note that for a fully deterministically-reproducible run,
            you must also limit the model to a single worker thread (`workers=1`), to eliminate ordering jitter
            from OS thread scheduling. (In Python 3, reproducibility between interpreter launches also requires
            use of the `PYTHONHASHSEED` environment variable to control hash randomization).
        max_vocab_size : int, optional
            Limits the RAM during vocabulary building; if there are more unique
            words than this, then prune the infrequent ones. Every 10 million word types need about 1GB of RAM.
            Set to `None` for no limit.
        max_final_vocab : int, optional
            Limits the vocab to a target vocab size by automatically picking a matching min_count. If the specified
            min_count is more than the calculated min_count, the specified min_count will be used.
            Set to `None` if not required.
        sample : float, optional
            The threshold for configuring which higher-frequency words are randomly downsampled,
            useful range is (0, 1e-5).
        hashfxn : function, optional
            Hash function to use to randomly initialize weights, for increased training reproducibility.
        iter : int, optional
            Number of iterations (epochs) over the corpus.
        trim_rule : function, optional
            Vocabulary trimming rule, specifies whether certain words should remain in the vocabulary,
            be trimmed away, or handled using the default (discard if word count < min_count).
            Can be None (min_count will be used, look to :func:`~gensim.utils.keep_vocab_item`),
            or a callable that accepts parameters (word, count, min_count) and returns either
            :attr:`gensim.utils.RULE_DISCARD`, :attr:`gensim.utils.RULE_KEEP` or :attr:`gensim.utils.RULE_DEFAULT`.
            The rule, if given, is only used to prune vocabulary during build_vocab() and is not stored as part of the
            model.

            The input parameters are of the following types:
                * `word` (str) - the word we are examining
                * `count` (int) - the word's frequency count in the corpus
                * `min_count` (int) - the minimum count threshold.
        sorted_vocab : {0, 1}, optional
            If 1, sort the vocabulary by descending frequency before assigning word indexes.
            See :meth:`~gensim.models.word2vec.Word2VecVocab.sort_vocab()`.
        batch_words : int, optional
            Target size (in words) for batches of examples passed to worker threads (and
            thus cython routines).(Larger batches will be passed if individual
            texts are longer than 10000 words, but the standard cython code truncates to that maximum.)
        compute_loss: bool, optional
            If True, computes and stores loss value which can be retrieved using
            :meth:`~gensim.models.word2vec.Word2Vec.get_latest_training_loss`.
        callbacks : iterable of :class:`~gensim.models.callbacks.CallbackAny2Vec`, optional
            Sequence of callbacks to be executed at specific stages during training.

        Returns:
            None

        Examples:
            >>> walker = DeepWalk()
            >>> result_list = [
            ...     ['Beijing', 'New York', 'Beijing', 'Shanghai', 'New York'],
            ...     ['London', 'Shanghai', 'Michigan', 'New York', 'Michigan'],
            ...     ['Zhuzhou', 'Beijing', 'New York', 'Shanghai', 'New York'],
            ...     ['Manchester', 'London', 'New York', 'Shanghai', 'Beijing'],
            ...     ['New York', 'Michigan', 'Shanghai', 'Zhuzhou', 'Shanghai'],
            ...     ['Michigan', 'New York', 'Beijing', 'New York', 'London'],
            ...     ['Shanghai', 'London', 'Manchester', 'London', 'Beijing']
            ... ]
            >>> walker._embedding(sentences=result_list, size=2, min_count=1, epochs=10000)
            >>> len(walker.model['Beijing'])
            2
        """

        self.model = Word2Vec(sentences=sentences, size=size, alpha=alpha, window=window, min_count=min_count,
                              max_vocab_size=max_vocab_size, sample=sample, seed=seed, workers=workers,
                              min_alpha=min_alpha,
                              sg=sg, hs=hs, negative=negative, ns_exponent=ns_exponent, cbow_mean=cbow_mean,
                              hashfxn=hashfxn, iter=epochs, null_word=null_word,
                              trim_rule=trim_rule, sorted_vocab=sorted_vocab, batch_words=batch_words,
                              compute_loss=compute_loss, callbacks=callbacks,
                              max_final_vocab=max_final_vocab)

    def train(self, walk_depth: int, walking_workers: int = 4, nodes_list: List[list] = None, size: int = 100,
              alpha: float = 0.025, window: int = 5, min_count: int = 1, max_vocab_size: int = None, sample=1e-3,
              seed=1, embedding_workers: int = 3, min_alpha: float = 0.0001, sg=0, hs=0, negative=5, ns_exponent=0.75,
              cbow_mean=1, hashfxn=hash, epochs: int = 5, null_word=0, trim_rule=None, sorted_vocab=1,
              batch_words: int = 10000, compute_loss=False, callbacks=(), max_final_vocab=None):
        """Function to train the DeepWalk model

        Args:
            See also docstring in concurrent_walk and _embedding functions.

        Returns:
            None

        Examples:
            >>> G = nx.Graph()
            >>> G.add_nodes_from([
            ...     ('Beijing', {'Country': 'CN'}),
            ...     ('London', {'Country': 'UK'}),
            ...     ('Zhuzhou', {'Country': 'CN'}),
            ...     ('Manchester', {'Country': 'UK'}),
            ...     ('New York', {'Country': 'US'}),
            ...     ('Michigan', {'Country': 'US'}),
            ...     ('Shanghai', {'Country': 'CN'})
            ... ])
            >>> G.add_edges_from([
            ...     ('Beijing', 'Zhuzhou'),
            ...     ('Beijing', 'London'),
            ...     ('Beijing', 'New York'),
            ...     ('Beijing', 'Shanghai'),
            ...     ('Zhuzhou', 'Shanghai'),
            ...     ('Shanghai', 'London'),
            ...     ('Shanghai', 'New York'),
            ...     ('Shanghai', 'Michigan'),
            ...     ('London', 'Manchester'),
            ...     ('London', 'New York'),
            ...     ('New York', 'Michigan')
            ... ])
            >>> walker = DeepWalk(G=G)
            >>> walker.train(walk_depth=5, size=2)
            >>> type(walker.model['Beijing']), len(walker.model['Zhuzhou'])
            (<class 'numpy.ndarray'>, 2)

        """
        # generate the walk path first
        self.walk_path = self.concurrent_walk(walk_depth=walk_depth, walking_workers=walking_workers,
                                              nodes_list=nodes_list)
        # then the model uses the walk path as 'sentences' to train a word2vec model
        self._embedding(sentences=self.walk_path, size=size, alpha=alpha, window=window, min_count=min_count,
                        max_vocab_size=max_vocab_size, sample=sample, seed=seed, workers=embedding_workers,
                        min_alpha=min_alpha, sg=sg, hs=hs, negative=negative, ns_exponent=ns_exponent,
                        cbow_mean=cbow_mean, hashfxn=hashfxn, epochs=epochs, null_word=null_word, trim_rule=trim_rule,
                        sorted_vocab=sorted_vocab, batch_words=batch_words, compute_loss=compute_loss,
                        callbacks=callbacks, max_final_vocab=max_final_vocab)

    def update_model(self, walk_path=None, walk_path_file=None, total_examples=None, total_words=None, epochs=None,
                     start_alpha=None, end_alpha=None, word_count=0, queue_factor=2, report_delay=1.0,
                     compute_loss=False, callbacks=()):
        """Function to update the word2vec model based on new walk path input.
        Args docstring is copied from gensim.models.Word2Vec

        Args:
        sentences : iterable of list of str
            The `sentences` iterable can be simply a list of lists of tokens, but for larger corpora,
            consider an iterable that streams the sentences directly from disk/network.
            See :class:`~gensim.models.word2vec.BrownCorpus`, :class:`~gensim.models.word2vec.Text8Corpus`
            or :class:`~gensim.models.word2vec.LineSentence` in :mod:`~gensim.models.word2vec` module for such examples.
            See also the `tutorial on data streaming in Python
            <https://rare-technologies.com/data-streaming-in-python-generators-iterators-iterables/>`_.
        corpus_file : str, optional
            Path to a corpus file in :class:`~gensim.models.word2vec.LineSentence` format.
            You may use this argument instead of `sentences` to get performance boost. Only one of `sentences` or
            `corpus_file` arguments need to be passed (not both of them).
        total_examples : int
            Count of sentences.
        total_words : int
            Count of raw words in sentences.
        epochs : int
            Number of iterations (epochs) over the corpus.
        start_alpha : float, optional
            Initial learning rate. If supplied, replaces the starting `alpha` from the constructor,
            for this one call to`train()`.
            Use only if making multiple calls to `train()`, when you want to manage the alpha learning-rate yourself
            (not recommended).
        end_alpha : float, optional
            Final learning rate. Drops linearly from `start_alpha`.
            If supplied, this replaces the final `min_alpha` from the constructor, for this one call to `train()`.
            Use only if making multiple calls to `train()`, when you want to manage the alpha learning-rate yourself
            (not recommended).
        word_count : int, optional
            Count of words already trained. Set this to 0 for the usual
            case of training on all words in sentences.
        queue_factor : int, optional
            Multiplier for size of queue (number of workers * queue_factor).
        report_delay : float, optional
            Seconds to wait before reporting progress.
        compute_loss: bool, optional
            If True, computes and stores loss value which can be retrieved using
            :meth:`~gensim.models.word2vec.Word2Vec.get_latest_training_loss`.
        callbacks : iterable of :class:`~gensim.models.callbacks.CallbackAny2Vec`, optional
            Sequence of callbacks to be executed at specific stages during training.

        Returns:
            None

        """
        self.model.train(sentences=walk_path, corpus_file=walk_path_file, total_examples=total_examples,
                         total_words=total_words, epochs=epochs, start_alpha=start_alpha, end_alpha=end_alpha,
                         word_count=word_count, queue_factor=queue_factor, report_delay=report_delay,
                         compute_loss=compute_loss, callbacks=callbacks)

    def vector(self, nodes: Union[List[str or int or float], Tuple[str or int or float], str, int, float],
               return_type: str = 'list') -> Union[list, dict]:
        """Function to return vectors of nodes in this model.

        Args:
            nodes: all nodes going to fetch the vector from this model
            return_type: can be either 'list' or 'dict' to decide the type of the returned result

        Returns:
            A list or dictionary including vectors for all these input nodes.

        Examples:
            >>> G = nx.Graph()
            >>> G.add_nodes_from([
            ...     ('Beijing', {'Country': 'CN'}),
            ...     ('London', {'Country': 'UK'}),
            ...     ('Zhuzhou', {'Country': 'CN'}),
            ...     ('Manchester', {'Country': 'UK'}),
            ...     ('New York', {'Country': 'US'}),
            ...     ('Michigan', {'Country': 'US'}),
            ...     ('Shanghai', {'Country': 'CN'})
            ... ])
            >>> G.add_edges_from([
            ...     ('Beijing', 'Zhuzhou'),
            ...     ('Beijing', 'London'),
            ...     ('Beijing', 'New York'),
            ...     ('Beijing', 'Shanghai'),
            ...     ('Zhuzhou', 'Shanghai'),
            ...     ('Shanghai', 'London'),
            ...     ('Shanghai', 'New York'),
            ...     ('Shanghai', 'Michigan'),
            ...     ('London', 'Manchester'),
            ...     ('London', 'New York'),
            ...     ('New York', 'Michigan')
            ... ])
            >>> walker = DeepWalk(G=G)
            >>> walker.train(walk_depth=5, size=2)
            >>> list_result = walker.vector('Beijing')
            >>> type(list_result), len(list_result)
            (<class 'list'>, 2)
            >>> dict_result = walker.vector(('Beijing', 'Zhuzhou'), return_type='dict')
            >>> dict_result.keys()
            dict_keys(['Beijing', 'Zhuzhou'])

        """
        # 'return_type' can be either 'list' or 'dict'
        assert return_type in ('list', 'dict'), SuiDeepWalkError(
            "The 'return_type' can be either 'list' or 'dict', got %s." % return_type)

        # if the 'nodes' is a list or tuple
        if isinstance(nodes, (list, tuple)):
            if return_type == 'list':
                result = []
                for node in nodes:
                    try:
                        # fetch the vector of this node from the model if it exists
                        result.append(list(self.model[node]))
                    except KeyError as keyerr:
                        raise SuiDeepWalkError(keyerr)
            else:
                result = {}
                for node in nodes:
                    try:
                        # fetch the vector of this node from the model if it exists
                        result.setdefault(node, list(self.model[node]))
                    except KeyError as keyerr:
                        raise SuiDeepWalkError(keyerr)
            return result

        else:
            try:
                # fetch the vector of this node from the model if it exists
                return list(self.model[nodes]) if return_type == 'list' else {nodes: list(self.model[nodes])}
            except KeyError as keyerr:
                raise SuiDeepWalkError(keyerr)

    def nearest_k(self, nodes_list: Union[list, tuple] = None, k: int = 30, negative: Union[list, tuple] = None,
                  restrict_vocab=None, indexer=None) -> dict:
        """Function to return top k nearest neighbours for every node in the input nodes_list.

        """
        if nodes_list is None:
            if len(self.G.nodes) == 0:
                raise ValueError("Parameter 'nodes_list' cannot be None if the length of G.nodes is 0.")

            nodes_list = self.G.nodes

        result = {}
        for node in nodes_list:
            result.setdefault(node, self.model.wv.most_similar(positive=node, negative=negative, topn=k,
                                                               restrict_vocab=restrict_vocab, indexer=indexer))
        return result



"""sui.toolbox
Tool kits for data processing
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from .top_k import top_k
from .build_series_data import build_session, random_walk



"""
    Module to load data to test models
    Date: 15/May/2020
    Author: Li Tang
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import pandas as pd
import os

__author__ = ['Li Tang']
__copyright__ = 'Li Tang'
__credits__ = ['Li Tang']
__license__ = 'MIT'
__version__ = '0.1.6'
__maintainer__ = ['Li Tang']
__email__ = 'litang1025@gmail.com'
__status__ = 'Production'


def iris():
    data_path = os.path.split(os.path.realpath(__file__))[0] + '/data/iris.csv'
    return pd.read_csv(data_path, header=None)


def titanic():
    data_path = os.path.split(os.path.realpath(__file__))[0] + '/data/titanic.csv'
    return pd.read_csv(data_path)


def movielens_1m(target='ratings'):
    data_path = os.path.split(os.path.realpath(__file__))[0] + '/data/movielens_1m/{}.csv'.format(target)
    return pd.read_csv(data_path)



from setuptools import setup

setup(name='sui',
      version='0.1.12',
      description='Some Useful Implementations',
      long_description='SUI(Some Useful Implementations) is a peckage including selected implementations apropos machine learning, deep learning, and reinforcement learning algorithms.',
      url='http://github.com/l-tang/sui',
      author='Li Tang',
      author_email='litang1025@gmail.com',
      license='MIT',
      packages=['sui', 'sui.ml', 'sui.dl', 'sui.rl', 'sui.toolbox', 'sui.ml.matrix_factorization', 'sui.graph'],
      zip_safe=False)
